
/*
 * Copyright (C) 2002 - 2007 Lehrstuhl Grafische Systeme, BTU Cottbus
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package de.grogra.ext.exchangegraph;

import java.util.HashMap;
import de.grogra.graph.impl.Edge;
import de.grogra.graph.impl.Node;
import de.grogra.xl.util.BidirectionalHashMap;
import de.grogra.imp3d.objects.*;
import de.grogra.ext.exchangegraph.nodes.*;
import de.grogra.rgg.Axiom;
import de.grogra.turtle.*;

/**
 * The IOContext contains informations about import and export the
 * exchange graph format.
 * <p>
 * If the same context is used for import <b>and</b> export you can
 * use the same data, eg. give the same GroIMP nodes the same XEG node id,
 * see {@link #nodeMap node map}. This is used for example in the
 * OpenAlea plugin.
 * 
 * @author Uwe Mannl
 *
 */
public class IOContext {
	
	/**
	 * Mapping of standard XEG node types to GroIMP node types.
	 */
	public static HashMap<String, String> importNodeTypes = new HashMap<String, String>();
	
	/**
	 * Mapping of GroIMP node types to standard XEG node types.
	 */
	public static HashMap<String, String> exportNodeTypes = new HashMap<String, String>();

	/**
	 * Mapping of GroIMP node types to XEG helper classes for import and export.
	 */
	@SuppressWarnings("unchecked")
	public static HashMap<Class, Class> xegNodeTypes = new HashMap<Class, Class>();
	
	static {
		importNodeTypes.put("Node", "de.grogra.imp3d.objects.ShadedNull");
		importNodeTypes.put("Box", "de.grogra.imp3d.objects.Box");
		importNodeTypes.put("Cone", "de.grogra.imp3d.objects.Cone");
		importNodeTypes.put("Sphere", "de.grogra.imp3d.objects.Sphere");
		importNodeTypes.put("Cylinder", "de.grogra.imp3d.objects.Cylinder");
		importNodeTypes.put("Frustum", "de.grogra.imp3d.objects.Frustum");
		importNodeTypes.put("F", "de.grogra.turtle.F");
		importNodeTypes.put("F0", "de.grogra.turtle.F0");
		importNodeTypes.put("M", "de.grogra.turtle.M");
		importNodeTypes.put("M0", "de.grogra.turtle.M0");
		importNodeTypes.put("RL", "de.grogra.turtle.RL");
		importNodeTypes.put("RU", "de.grogra.turtle.RU");
		importNodeTypes.put("RH", "de.grogra.turtle.RH");
		importNodeTypes.put("RV", "de.grogra.turtle.RV");
		importNodeTypes.put("RG", "de.grogra.turtle.RG");
		importNodeTypes.put("AdjustLU", "de.grogra.turtle.AdjustLU");
		importNodeTypes.put("L", "de.grogra.turtle.L");
		importNodeTypes.put("D", "de.grogra.turtle.D");
		importNodeTypes.put("P", "de.grogra.turtle.P");
		importNodeTypes.put("Axiom", "de.grogra.rgg.Axiom");
		importNodeTypes.put("Mesh", "de.grogra.imp3d.objects.MeshNode");
		
		xegNodeTypes.put(Node.class, XEGNode.class);
		xegNodeTypes.put(Null.class, XEGNode.class);
		xegNodeTypes.put(ShadedNull.class, XEGNode.class);
		xegNodeTypes.put(Box.class, XEGBox.class);
		xegNodeTypes.put(Cone.class, XEGCone.class);
		xegNodeTypes.put(Sphere.class, XEGSphere.class);
		xegNodeTypes.put(Cylinder.class, XEGCylinder.class);
		xegNodeTypes.put(Frustum.class, XEGFrustum.class);
		xegNodeTypes.put(F.class, XEGF.class);
		xegNodeTypes.put(F0.class, XEGF0.class);
		xegNodeTypes.put(M.class, XEGM.class);
		xegNodeTypes.put(M0.class, XEGM0.class);
		xegNodeTypes.put(RL.class, XEGRL.class);
		xegNodeTypes.put(RU.class, XEGRU.class);
		xegNodeTypes.put(RH.class, XEGRH.class);
		xegNodeTypes.put(RV.class, XEGRV.class);
		xegNodeTypes.put(RG.class, XEGRG.class);
		xegNodeTypes.put(AdjustLU.class, XEGAdjustLU.class);
		xegNodeTypes.put(L.class, XEGL.class);
		xegNodeTypes.put(D.class, XEGD.class);
		xegNodeTypes.put(P.class, XEGP.class);
		xegNodeTypes.put(Axiom.class, XEGAxiom.class);
		xegNodeTypes.put(MeshNode.class, XEGMeshNode.class);
				
		exportNodeTypes.put("de.grogra.graph.impl.Node", "Node");
		exportNodeTypes.put("de.grogra.imp3d.objects.Null", "Node");
		exportNodeTypes.put("de.grogra.imp3d.objects.ShadedNull", "Node");
		exportNodeTypes.put("de.grogra.imp3d.objects.Box", "Box");
		exportNodeTypes.put("de.grogra.imp3d.objects.Cone", "Cone");
		exportNodeTypes.put("de.grogra.imp3d.objects.Sphere", "Sphere");
		exportNodeTypes.put("de.grogra.imp3d.objects.Cylinder", "Cylinder");
		exportNodeTypes.put("de.grogra.imp3d.objects.Frustum", "Frustum");
		exportNodeTypes.put("de.grogra.turtle.F", "F");
		exportNodeTypes.put("de.grogra.turtle.F0", "F0");
		exportNodeTypes.put("de.grogra.turtle.M", "M");
		exportNodeTypes.put("de.grogra.turtle.M0", "M0");
		exportNodeTypes.put("de.grogra.turtle.RL", "RL");
		exportNodeTypes.put("de.grogra.turtle.RU", "RU");
		exportNodeTypes.put("de.grogra.turtle.RH", "RH");
		exportNodeTypes.put("de.grogra.turtle.RV", "RV");
		exportNodeTypes.put("de.grogra.turtle.RG", "RG");
		exportNodeTypes.put("de.grogra.turtle.AdjustLU", "AdjustLU");
		exportNodeTypes.put("de.grogra.turtle.L", "L");
		exportNodeTypes.put("de.grogra.turtle.D", "D");
		exportNodeTypes.put("de.grogra.turtle.P", "P");
		exportNodeTypes.put("de.grogra.rgg.Axiom", "Axiom");
		exportNodeTypes.put("de.grogra.imp3d.objects.MeshNode", "Mesh");
	}
	
	/**
	 * Map contains the link between XEG node ids and the corresponding GroIMP
	 * nodes. It is used to set the same ids for export like for import for same nodes. 
	 */
	private BidirectionalHashMap<Long, Node> nodeMap = new BidirectionalHashMap<Long, Node>();
	
	/**
	 * Map contains the link between XEG edge ids and the corresponding GroIMP
	 * edges. It is used to set the same ids for export like for import for same edges. 
	 */
	private BidirectionalHashMap<Long, Edge> edgeMap = new BidirectionalHashMap<Long, Edge>();
	
	/**
	 * Map contains specific edge types read for export which were read
	 * during import. Can be empty if no import was done. 
	 */
	private BidirectionalHashMap<Integer, String> edgeTypes = new BidirectionalHashMap<Integer, String>();
	
	/**
	 * Returns the {@link #nodeMap node map}.
	 * @return
	 */
	public BidirectionalHashMap<Long, Node> getNodeMap() {
		return nodeMap;
	}

	/**
	 * Returns the {@link #edgeMap edge map}.
	 * @return edge map
	 */
	public BidirectionalHashMap<Long, Edge> getEdgeMap() {
		return edgeMap;
	}

	/**
	 * Returns the {@link #edgeTypes edge types map}.
	 * @return edge types map
	 */
	public BidirectionalHashMap<Integer, String> getEdgeTypes() {
		return edgeTypes;
	}
	
}
